# frozen_string_literal: true

require "http/2/version"

module HTTP2
  EMPTY = [].freeze
end

require "http/2/extensions"
require "http/2/base64"
require "http/2/error"
require "http/2/emitter"
require "http/2/flow_buffer"
require "http/2/header"
require "http/2/framer"
require "http/2/connection"
require "http/2/client"
require "http/2/server"
require "http/2/stream"
